/**
 * The Pagination Utility - The Pagination widget provides a set of controls to navigate through paged data.
 *
 * @module aui-pagination
 */

var Lang = A.Lang,
    isBoolean = Lang.isBoolean,
    isNumber = Lang.isNumber,

    ACTIVE = 'active',
    BOUNDING_BOX = 'boundingBox',
    CHANGE_REQUEST = 'changeRequest',
    CIRCULAR = 'circular',
    CLICK = 'click',
    CONTENT_BOX = 'contentBox',
    DISABLED = 'disabled',
    ITEMS = 'items',
    LI = 'li',
    NEXT = 'next',
    OFFSET = 'offset',
    PAGE = 'page',
    PAGINATION = 'pagination',
    PREV = 'prev',
    RENDERED = 'rendered',
    TOTAL = 'total',

    getCN = A.getClassName,

    CSS_ACTIVE = getCN(ACTIVE),
    CSS_DISABLED = getCN(DISABLED);

    /**
     * A base class for Pagination, providing:
     * <ul>
     *    <li>Widget Lifecycle (initializer, renderUI, bindUI, syncUI, destructor)</li>
     *    <li>Set of controls to navigate through paged data</li>
     * </ul>
     *
     * Quick Example:<br/>
     *
     * <pre><code>var instance = new A.Pagination({
     *  boundingBox: '#pagination',
     *  circular: true,
     *  total: 10
     * }).render();
     * </code></pre>
     *
     * Check the list of <a href="Pagination.html#configattributes">Configuration Attributes</a> available for
     * Pagination.
     *
     * @param config {Object} Object literal specifying widget configuration properties.
     *
     * @class Pagination
     * @constructor
     * @extends Widget
     */
var Pagination = A.Component.create(
    {
        /**
         * Static property provides a string to identify the class.
         *
         * @property Paginator.NAME
         * @type String
         * @static
         */
        NAME: PAGINATION,

        /**
         * Static property used to define the default attribute
         * configuration for the Paginator.
         *
         * @property Paginator.ATTRS
         * @type Object
         * @static
         */
        ATTRS: {
            /**
             * When enabled this property allows the navigation to go back to
             * the beggining when it reaches the last page, the opposite behavior
             * is also true. Incremental page navigation could happen clicking the
             * control arrows or invoking <code>.next()</code> and
             * <code>.prev()</code> methods.
             *
             * @attribute circular
             * @default true
             * @type boolean
             */
            circular: {
                validator: isBoolean,
                value: true
            },

            /**
             * Holds the page items as a NodeList. The list could be queried
             * from the DOM trough Widget HTML_PARSER or generated if
             * <a href="Paginator.html#config_total">total</a> is specified.
             *
             * @attribute items
             * @default undefined
             * @type NodeList
             */
            items: {
            },

            /**
             * Initial page offset.
             *
             * @attribute offset
             * @default 0
             * @type Number
             */
            offset: {
                setter: '_setInt',
                value: 1
            },

            /**
             * Page to display on initial paint.
             *
             * @attribute page
             * @default 0
             * @type Number
             */
            page: {
                setter: '_setInt',
                value: 0
            },

            /**
             * Total number of page links available. If set, the new
             * <a href="Paginator.html#config_items">items</a> node list will
             * be rendered.
             *
             * @attribute total
             * @default 0
             * @type Number
             */
            total: {
                setter: '_setInt',
                value: 0
            },

            strings: {
                value: {
                    next: 'Next',
                    prev: 'Prev'
                }
            }
        },

        HTML_PARSER: {
            items: function(srcNode) {
                return this._queryItemsIfNotSet(srcNode);
            },
            total: function() {
                return this._countItemsInDoc();
            }
        },

        UI_ATTRS: [OFFSET, PAGE, TOTAL],

        prototype: {
            CONTENT_TEMPLATE: '<ul></ul>',
            ITEM_TEMPLATE: '<li><a href="#">{content}</a></li>',
            TOTAL_CONTROLS: 2,

            items: null,
            lastState: null,

            syncUI: function() {
                var instance = this,
                    page = instance.get(PAGE);

                if (page > 0) {
                    instance._dispatchRequest({page:page});
                }
            },

            bindUI: function() {
                var instance = this,
                    boundingBox = instance.get(BOUNDING_BOX);

                instance.on('pageChange', instance._onPageChange);
                instance.publish(CHANGE_REQUEST, {
                    defaultFn: instance._defChangeRequest
                });
                boundingBox.delegate(CLICK, instance._onClickItem, LI, instance);
            },

            renderUI: function() {
                var instance = this;

                instance._renderItemsUI(instance.get(TOTAL));
            },

            /**
             * Retrieve the item node from the passesed item index parameter.
             * If passed item is a node instead of the index returns itself.
             *
             * @method getItem
             * @param {Node | Number} i Index or Node reference.
             * @return {Node}
             */
            getItem: function(i) {
                var instance = this;

                if (isNumber(i)) {
                    var items = instance.get(ITEMS);
                    if (items) {
                        i = items.item(i);
                    }
                }

                return i;
            },

            /**
             * Retrieve page number inclusing offset e.g., if offset
             * is 100 and active page is 5, this method returns 105.
             *
             * @method getOffsetPageNumber
             * @return {Number}
             */
            getOffsetPageNumber: function() {
                var instance = this;

                return instance.get(OFFSET) + instance.get(PAGE);
            },

            /**
             * Retrieve total number of pages including offset e.g., if offset
             * is 100 and total 10, this method returns 110.
             *
             * @method getOffsetTotalPages
             * @return {Number}
             */
            getOffsetTotalPages: function() {
                var instance = this;

                return instance.get(OFFSET) + instance.get(TOTAL);
            },

            /**
             * Retrieve total number of dom items representing the links,
             * including the arrow control items. Do not include the offset.
             *
             * @method getTotalItems
             * @return {Number}
             */
            getTotalItems: function() {
                var instance = this;

                return instance.get(TOTAL) + instance.TOTAL_CONTROLS;
            },

            /**
             * Navigate to the next page.
             *
             * @method next
             */
            next: function() {
                var instance = this,
                    total = instance.get(TOTAL);

                if (total === 0) {
                    return;
                }

                var page = instance.get(PAGE);

                instance._dispatchRequest({
                    page: (instance.get(CIRCULAR) && (page === total)) ?
                            1 :
                            Math.min(total, ++page)
                });

                return instance;
            },

            /**
             * Navigate to the previous page.
             *
             * @method prev
             */
            prev: function() {
                var instance = this,
                    total = instance.get(TOTAL);

                if (total === 0) {
                    return;
                }

                var page = instance.get(PAGE);

                instance._dispatchRequest({
                    page: (instance.get(CIRCULAR) && (page === 1)) ?
                            total :
                            Math.max(1, --page)
                });

                return instance;
            },

            /**
             * Set the new pagination state. The state is a payload object
             * containing the page number, e.g. <code>{page:1}</code>.
             *
             * @method setState
             */
            setState: function(state) {
                var instance = this;

                instance.set(PAGE, state.page);
                instance.lastState = state;
            },

            _countItemsInDoc: function() {
                var instance = this,
                    srcNode = instance.get('srcNode');

                return Math.max(0, instance._queryItemsIfNotSet(srcNode).size() - instance.TOTAL_CONTROLS);
            },

            _defChangeRequest: function(event) {
                var instance = this;

                instance.setState(event.state);
            },

            _dispatchRequest: function(state) {
                var instance = this;

                instance.fire(CHANGE_REQUEST, {
                    lastState: instance.lastState,
                    state: state
                });
            },

            _queryItemsIfNotSet: function(srcNode) {
                var instance = this;

                if (!instance.items) {
                    instance.items = srcNode.all(LI);
                }
                return instance.items;
            },

            _onClickItem: function(event) {
                var instance = this,
                    item = event.currentTarget;

                event.preventDefault();

                if (item.hasClass(CSS_DISABLED) || item.hasClass(CSS_ACTIVE)) {
                    return;
                }

                var items = instance.get(ITEMS),
                    index = items.indexOf(item),
                    lastIndex = items.size() - 1;

                switch(index) {
                    case 0:
                        instance.prev();
                        break;
                    case lastIndex:
                        instance.next();
                        break;
                    default:
                        instance._dispatchRequest({page:index});
                        break;
                }
            },

            _onPageChange: function(event) {
                var instance = this;

                if (event.prevVal !== event.newVal) {
                    var item = instance.getItem(event.prevVal);
                    if (item) {
                        item.removeClass(CSS_ACTIVE);
                    }
                }
            },

            _renderItemsUI: function(total) {
                var instance = this;

                // Do not render anything if the number of markup items matches
                // with the passed total attributes.
                if ((total > 0) && (total === instance._countItemsInDoc())) {
                    return;
                }

                var tpl = instance.ITEM_TEMPLATE,
                    offset = instance.get(OFFSET),
                    i,
                    buffer = '';

                buffer += Lang.sub(tpl, {
                    content: instance.getString(PREV)
                });

                for (i = offset + 1; i <= (offset + total); i++) {
                    buffer += Lang.sub(tpl, { content: i });
                }

                buffer += Lang.sub(tpl, {
                    content: instance.getString(NEXT)
                });

                var items = A.NodeList.create(buffer);
                instance.set(ITEMS, items);
                instance.get(CONTENT_BOX).setContent(items);
            },

            _setInt: function(val) {
                return Lang.toInt(val);
            },

            _syncNavigationUI: function() {
                var instance = this,
                    items = instance.get(ITEMS);

                items.first().toggleClass(
                    CSS_DISABLED, instance.get(PAGE) === 1);

                items.last().toggleClass(
                    CSS_DISABLED, instance.get(PAGE) === instance.get(TOTAL));
            },

            _uiSetOffset: function(val) {
                var instance = this;

                if (instance.get(RENDERED)) {
                    instance._renderItemsUI(instance.get(TOTAL));
                }
            },

            _uiSetPage: function(val) {
                var instance = this;

                if (!instance.get(CIRCULAR)) {
                    instance._syncNavigationUI();
                }

                // Do not activate first and last items, they are used for controls.
                if (val === 0 || val === instance.getTotalItems()) {
                    return;
                }

                var item = instance.getItem(val);

                if (item) {
                    item.addClass(CSS_ACTIVE);
                }
            },

            _uiSetTotal: function(val) {
                var instance = this;

                if (instance.get(RENDERED)) {
                    instance._renderItemsUI(val);
                }
            }
        }
    }
);

A.Pagination = Pagination;